/* clock.c -- (C) Geoffrey Reynolds, May 2006.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include "sr5sieve.h"
#include "config.h"


/* Return the CPU time used by this process to date, in milliseconds. Using
   32 bits for millisecond times gives 49 days before wrap-around.
*/
#if HAVE_GETRUSAGE
#include <sys/resource.h>
uint32_t millisec_clock(void)
{
  int ret;
  struct rusage r;

  ret = getrusage(RUSAGE_SELF, &r);

  assert(ret == 0);

  return (r.ru_utime.tv_sec + r.ru_stime.tv_sec) * 1000
    + (r.ru_utime.tv_usec + r.ru_stime.tv_usec) / 1000;
}

uint32_t sec_clock(void)
{
  int ret;
  struct rusage r;

  ret = getrusage(RUSAGE_SELF, &r);

  assert(ret == 0);

  return (r.ru_utime.tv_sec + r.ru_stime.tv_sec);
}
#elif defined(_WIN32)
/* clock() doesn't seem to work in Windows, it just returns the elapsed time
   since the process started. That might be a mingw32 bug, but no harm using
   this code for other Windows compilers just in case.
*/
#include <windows.h>

/* Return total CPU time spent on this process in units of 100 nanosec.
 */
static uint64_t ns100_clock(void)
{
  FILETIME ft_create, ft_exit, ft_kernel, ft_user;
  ULARGE_INTEGER ns100_kernel, ns100_user;

  GetProcessTimes(GetCurrentProcess(),&ft_create,&ft_exit,&ft_kernel,&ft_user);

  ns100_kernel.u.LowPart = ft_kernel.dwLowDateTime;
  ns100_kernel.u.HighPart = ft_kernel.dwHighDateTime;
  ns100_user.u.LowPart = ft_user.dwLowDateTime;
  ns100_user.u.HighPart = ft_user.dwHighDateTime;

  return (ns100_kernel.QuadPart + ns100_user.QuadPart);
}

uint32_t millisec_clock(void)
{
  return ns100_clock()/10000;
}

uint32_t sec_clock(void)
{
  return ns100_clock()/10000000;
}
#else
/* If clock_t is an unsigned 32 bit type and CLOCKS_PER_SEC is 1 million
   (typical case) then clock() will wrap around every 72 minutes. But as
   long as this function is called at least once every 72 minutes, and not
   called from a signal handler, it will work correctly. We can ensure this
   happens by setting the maximum reporting period to 1 hour in srsieve.h.
*/
#include <time.h>

static uint64_t clocks(void)
{
  static clock_t last_clock = 0;
  static uint64_t accumulated_clocks = 0;
  clock_t this_clock = clock();

  assert(this_clock != (clock_t)-1);

  accumulated_clocks += (this_clock - last_clock);
  last_clock = this_clock;

  return accumulated_clocks;
}

uint32_t millisec_clock(void)
{
  return clocks() * 1000 / CLOCKS_PER_SEC;
}

uint32_t sec_clock(void)
{
  return clocks() / CLOCKS_PER_SEC;
}
#endif /* HAVE_RUSAGE */


/* Accumulated CPU time in seconds, accurate to 1 millisec.
 */
static double acc_seconds;
double get_accumulated_cpu(void)
{
  return acc_seconds + (double)millisec_clock()/1000;
}

void set_accumulated_cpu(double seconds)
{
  acc_seconds = seconds - (double)millisec_clock()/1000;
}

/* For benchmarking purposes.
 */
#if USE_ASM && defined(__GNUC__)
#if defined(__i386__)
uint64_t timestamp(void)
{
  uint64_t ret;
  asm volatile ("xor     %%eax, %%eax"  "\n\t"
                "cpuid"                 "\n\t"
                "rdtsc"
                : "=A" (ret) : : "%ebx", "%ecx", "cc" );
  return ret;
}
#elif defined(__x86_64__)
uint64_t timestamp(void)
{
  uint64_t ret;
  asm volatile ("xor     %%eax, %%eax"  "\n\t"
                "cpuid"                 "\n\t"
                "rdtsc"                 "\n\t"
                "shl     $32, %%rdx"    "\n\t"
                "or      %%rdx, %%rax"
                : "=a" (ret) : : "%rbx", "%rcx", "%rdx", "cc" );
  return ret;
}
#elif defined(__ppc64__) || defined(__powerpc64__)
uint64_t timestamp(void)
{
  uint64_t ret;
  asm volatile ("mftb %0" : "=r" (ret) );
  return ret;
}
#elif HAVE_GETTIMEOFDAY
#include <stdlib.h>
#include <sys/time.h>
uint64_t timestamp(void)
{
  struct timeval thistime;
  gettimeofday(&thistime, NULL);
  return thistime.tv_sec * 1000000 + thistime.tv_usec;
}
#else
#include <time.h>
uint64_t timestamp(void)
{
  return clock();
}
#endif
#endif
