/* mulmod-k8.h -- (C) Geoffrey Reynolds, July 2006.

   Inline modular multiplication routines for x86-64/SSE2.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#if defined(NEED_mulmod64) && !defined(HAVE_mulmod64)
#define HAVE_mulmod64

#define SSE_MODE_BITS "(0x6000)"  /* Round To Zero */

#define MOD64_MAX_BITS 51

/* Saved copy of SSE rounding mode bits. */
extern uint16_t mod64_rnd;

extern double one_over_p attribute ((aligned(16)));
extern double b_over_p attribute ((aligned(16)));

/* a*b (mod p), where a,b < p < 2^51.
   Assumes SSE operations round to zero.
   Assumes one_over_p = 1.0/p computed in round-to-zero mode.
*/
static inline uint64_t mulmod64(uint64_t a, uint64_t b, uint64_t p)
{
  register uint64_t ret, tmp;
  register double x, y;

  asm ("cvtsi2sdq %0, %2"    "\n\t"
       "cvtsi2sdq %5, %3"    "\n\t"
       "mulsd     %7, %2"    "\n\t"
       "mulsd     %2, %3"    "\n\t"
       "cvtsd2siq %3, %1"    "\n\t"
       "imul      %5, %0"    "\n\t"
       "imul      %6, %1"    "\n\t"
       "sub       %1, %0"    "\n\t"
       "mov       %0, %1"    "\n\t"
       "sub       %6, %1"    "\n\t"
       "jl        0f"        "\n\t"
       "mov       %1, %0"    "\n"
       "0:"
       : "=r" (ret), "=&r" (tmp), "=&x" (x), "=&x" (y)
       : "0" (a), "rm" (b), "rm" (p), "xm" (one_over_p), "m" (mod64_rnd)
       : "cc" );

  return ret;
}

/* Set SSE rounding mode to round-to-zero.
   Set one_over_p to 1.0/p computed in round-to-zero mode.
   Saves the old rounding mode in mod64_rnd.
*/
static inline void mod64_init(uint64_t p)
{
  uint32_t tmp;
  register double x;

  asm volatile ("stmxcsr   %0"              "\n\t"
                "mov       %0, %%ax"        "\n\t"
                "and       $"SSE_MODE_BITS", %%ax"    "\n\t"
                "orw       $"SSE_MODE_BITS", %0"      "\n\t"
                "mov       %%ax, %1"        "\n\t"  /* Save old bits */
                "ldmxcsr   %0"              "\n\t"
                "mov       $1, %%eax"       "\n\t"
                "cvtsi2sdq %4, %3"          "\n\t"
                "cvtsi2sd  %%eax, %2"       "\n\t"
                "divsd     %3, %2"

                : "=m" (tmp), "+m" (mod64_rnd), "=x" (one_over_p), "=x" (x)
                : "rm" (p)
                : "%rax", "cc" );
}

/* Restore the previous SSE rounding mode from mod64_rnd.
*/
static inline void mod64_fini(void)
{
  uint32_t tmp;

  asm volatile ("stmxcsr   %0"              "\n\t"
                "mov       %0, %%ax"        "\n\t"
                "and       $~"SSE_MODE_BITS", %%ax"    "\n\t"
                "or        %1, %%ax"        "\n\t"  /* Restore old bits */
                "mov       %%ax, %0"        "\n\t"
                "ldmxcsr   %0"

                : "=m" (tmp), "+m" (mod64_rnd)
                : 
                : "%rax", "cc" );
}

/* a*b (mod p), where a,b < p < 2^51.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define PRE2_MULMOD64(a,b,p) ({ \
  register uint64_t ret, tmp; \
  register double x; \
  asm ("cvtsi2sdq %0, %2"    "\n\t" \
       "mulsd     %6, %2"    "\n\t" \
       "cvtsd2siq %2, %1"    "\n\t" \
       "imul      %4, %0"    "\n\t" \
       "imul      %5, %1"    "\n\t" \
       "sub       %1, %0"    "\n\t" \
       "mov       %0, %1"    "\n\t" \
       "sub       %5, %1"    "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %1, %0"    "\n" \
       "0:" \
       : "=r" (ret), "=&r" (tmp), "=&x" (x) \
       : "0" (a), "r" (b), "r" (p), "x" (b_over_p), "m" (mod64_rnd) \
       : "cc" ); \
  ret; })

/* Set b_over_p to b/p, computed in round-to-zero mode.
   Assumes SSE operations round to zero.
   Assumes one_over_p = 1.0/p computed in round-to-zero mode.
 */
#define PRE2_MULMOD64_INIT(b) \
  asm ("cvtsi2sdq %2, %0"    "\n\t" \
       "mulsd     %3, %0" \
       : "=&x" (b_over_p), "+m" (mod64_rnd) \
       : "rm" ((uint64_t)b), "xm" (one_over_p) \
       : "cc" )

#define PRE2_MULMOD64_FINI() {}

#endif /* defined(NEED_mulmod64) && !defined(HAVE_mulmod64) */


#if defined(NEED_sqrmod64) && !defined(HAVE_sqrmod64)
#define HAVE_sqrmod64
extern uint16_t mod64_rnd;

/* b^2 (mod p), where b < p < 2^51.
   Assumes SSE operations round to zero.
   Assumes one_over_p = 1.0/p computed in round-to-zero mode.
*/
static inline uint64_t sqrmod64(uint64_t b, uint64_t p)
{
  register uint64_t ret, tmp;
  register double x;

  asm ("cvtsi2sdq %0, %2"    "\n\t"
       "mulsd     %2, %2"    "\n\t"
       "mulsd     %5, %2"    "\n\t"
       "cvtsd2siq %2, %1"    "\n\t"
       "imul      %0, %0"    "\n\t"
       "imul      %4, %1"    "\n\t"
       "sub       %1, %0"    "\n\t"
       "mov       %0, %1"    "\n\t"
       "sub       %4, %1"    "\n\t"
       "jl        0f"        "\n\t"
       "mov       %1, %0"    "\n"
       "0:"
       : "=r" (ret), "=&r" (tmp), "=&x" (x)
       : "0" (b), "rm" (p), "xm" (one_over_p), "m" (mod64_rnd)
       : "cc" );

  return ret;
}
#endif /* defined(NEED_sqrmod64) && !defined(HAVE_sqrmod64) */


#if defined(NEED_powmod64) && !defined(HAVE_powmod64)
#define HAVE_powmod64
/*
  b^n (mod p)
*/
uint64_t powmod64_k8(uint64_t b, uint64_t n, uint64_t p, double invp)
     attribute ((const));
#define powmod64(b,n,p) powmod64_k8(b,n,p,one_over_p)
#endif


#if defined(NEED_vector) && !defined(HAVE_vector)
#define HAVE_vector

/* Given uint64_t X[2], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i+2] <-- X[i] * b (mod p), for 0 <= i < 2.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC2_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  register double x0, x1; \
 \
  asm ("mov       %11, %4"   "\n\t" \
       "mov       %12, %5"   "\n\t" \
       "cvtsi2sdq %4, %2"    "\n\t" \
       "cvtsi2sdq %5, %3"    "\n\t" \
       "mulsd     %10, %2"   "\n\t" \
       "mulsd     %10, %3"   "\n\t" \
       "cvtsd2siq %2, %6"    "\n\t" \
       "cvtsd2siq %3, %7"    "\n\t" \
 \
       "imul      %8, %4"    "\n\t" \
       "imul      %8, %5"    "\n\t" \
       "imul      %9, %6"    "\n\t" \
       "imul      %9, %7"    "\n\t" \
       "sub       %6, %4"    "\n\t" \
       "sub       %7, %5"    "\n\t" \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "sub       %9, %4"    "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %9, %5"    "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[2]), "=m" ((X)[3]), "=&x" (x0), "=&x" (x1), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" ((X)[0]), "m" ((X)[1]), \
         "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[2], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 2.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC2_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  register double x0, x1; \
 \
  asm ("mov       %0, %4"    "\n\t" \
       "mov       %1, %5"    "\n\t" \
       "cvtsi2sdq %4, %2"    "\n\t" \
       "cvtsi2sdq %5, %3"    "\n\t" \
       "mulsd     %10, %2"   "\n\t" \
       "mulsd     %10, %3"   "\n\t" \
       "cvtsd2siq %2, %6"    "\n\t" \
       "cvtsd2siq %3, %7"    "\n\t" \
 \
       "imul      %8, %4"    "\n\t" \
       "imul      %8, %5"    "\n\t" \
       "imul      %9, %6"    "\n\t" \
       "imul      %9, %7"    "\n\t" \
       "sub       %6, %4"    "\n\t" \
       "sub       %7, %5"    "\n\t" \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "sub       %9, %4"    "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %9, %5"    "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                         \
       : "+m" ((X)[0]), "+m" ((X)[1]), "=&x" (x0), "=&x" (x1), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[4], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i+4] <-- X[i] * b (mod p), for 0 <= i < 4.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC4_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  register double x0, x1; \
 \
  asm ("mov       %11, %4"   "\n\t" \
       "mov       %12, %5"   "\n\t" \
       "cvtsi2sdq %4, %2"    "\n\t" \
       "cvtsi2sdq %5, %3"    "\n\t" \
       "mulsd     %10, %2"   "\n\t" \
       "mulsd     %10, %3"   "\n\t" \
       "cvtsd2siq %2, %6"    "\n\t" \
       "cvtsd2siq %3, %7"    "\n\t" \
 \
       "imul      %8, %4"    "\n\t" \
       "imul      %8, %5"    "\n\t" \
       "imul      %9, %6"    "\n\t" \
       "imul      %9, %7"    "\n\t" \
       "sub       %6, %4"    "\n\t" \
       "sub       %7, %5"    "\n\t" \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "sub       %9, %4"    "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %9, %5"    "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[4]), "=m" ((X)[5]), "=&x" (x0), "=&x" (x1), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" ((X)[0]), "m" ((X)[1]), \
         "m" (mod64_rnd) \
       : "cc" ); \
 \
  asm ("mov       %11, %4"   "\n\t" \
       "mov       %12, %5"   "\n\t" \
       "cvtsi2sdq %4, %2"    "\n\t" \
       "cvtsi2sdq %5, %3"    "\n\t" \
       "mulsd     %10, %2"   "\n\t" \
       "mulsd     %10, %3"   "\n\t" \
       "imul      %8, %4"    "\n\t" \
       "imul      %8, %5"    "\n\t" \
       "cvtsd2siq %2, %6"    "\n\t" \
       "cvtsd2siq %3, %7"    "\n\t" \
 \
       "imul      %9, %6"    "\n\t" \
       "imul      %9, %7"    "\n\t" \
       "sub       %6, %4"    "\n\t" \
       "sub       %7, %5"    "\n\t" \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "sub       %9, %4"    "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %9, %5"    "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[6]), "=m" ((X)[7]), "=&x" (x0), "=&x" (x1), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" ((X)[2]), "m" ((X)[3]), \
         "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[4], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 4.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC4_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  register double x0, x1, x2, x3; \
 \
  asm ("mov       %0, %8"    "\n\t" \
       "mov       %1, %9"    "\n\t" \
       "mov       %2, %10"   "\n\t" \
       "mov       %3, %11"   "\n\t" \
       "cvtsi2sdq %8, %4"    "\n\t" \
       "cvtsi2sdq %9, %5"    "\n\t" \
       "cvtsi2sdq %10, %6"   "\n\t" \
       "cvtsi2sdq %11, %7"   "\n\t" \
       "mulsd     %18, %4"   "\n\t" \
       "mulsd     %18, %5"   "\n\t" \
       "mulsd     %18, %6"   "\n\t" \
       "mulsd     %18, %7"   "\n\t" \
       "imul      %16, %8"   "\n\t" \
       "imul      %16, %9"   "\n\t" \
       "imul      %16, %10"  "\n\t" \
       "imul      %16, %11"  "\n\t" \
       "cvtsd2siq %4, %12"   "\n\t" \
       "cvtsd2siq %5, %13"   "\n\t" \
       "cvtsd2siq %6, %14"   "\n\t" \
       "cvtsd2siq %7, %15"   "\n\t" \
 \
       "imul      %17, %12"  "\n\t" \
       "imul      %17, %13"  "\n\t" \
       "imul      %17, %14"  "\n\t" \
       "imul      %17, %15"  "\n\t" \
       "sub       %12, %8"   "\n\t" \
       "sub       %13, %9"   "\n\t" \
       "sub       %14, %10"  "\n\t" \
       "sub       %15, %11"  "\n\t" \
 \
       "mov       %8, %0"    "\n\t" \
       "mov       %9, %1"    "\n\t" \
       "mov       %10, %2"   "\n\t" \
       "mov       %11, %3"   "\n\t" \
       "sub       %17, %8"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %8, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %17, %9"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %9, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %17, %10"  "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %10, %2"   "\n"   \
       "2:"                  "\t"   \
       "sub       %17, %11"  "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %11, %3"   "\n"   \
       "3:"                         \
       : "+m" ((X)[0]), "+m" ((X)[1]), "+m" ((X)[2]), "+m" ((X)[3]), \
         "=&x" (x0), "=&x" (x1), "=&x" (x2), "=&x" (x3), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[8], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i+8] <-- X[i] * b (mod p), for 0 <= i < 8.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC8_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  register double x0, x1, x2, x3; \
 \
  asm ("mov       %19, %8"   "\n\t" \
       "mov       %20, %9"   "\n\t" \
       "mov       %21, %10"  "\n\t" \
       "mov       %22, %11"  "\n\t" \
       "cvtsi2sdq %8,  %4"   "\n\t" \
       "cvtsi2sdq %9,  %5"   "\n\t" \
       "cvtsi2sdq %10, %6"   "\n\t" \
       "cvtsi2sdq %11, %7"   "\n\t" \
       "mulsd     %18, %4"   "\n\t" \
       "mulsd     %18, %5"   "\n\t" \
       "mulsd     %18, %6"   "\n\t" \
       "mulsd     %18, %7"   "\n\t" \
       "imul      %16, %8"   "\n\t" \
       "imul      %16, %9"   "\n\t" \
       "imul      %16, %10"  "\n\t" \
       "imul      %16, %11"  "\n\t" \
       "cvtsd2siq %4, %12"   "\n\t" \
       "cvtsd2siq %5, %13"   "\n\t" \
       "cvtsd2siq %6, %14"   "\n\t" \
       "cvtsd2siq %7, %15"   "\n\t" \
 \
       "imul      %17, %12"  "\n\t" \
       "imul      %17, %13"  "\n\t" \
       "imul      %17, %14"  "\n\t" \
       "imul      %17, %15"  "\n\t" \
       "sub       %12, %8"   "\n\t" \
       "sub       %13, %9"   "\n\t" \
       "sub       %14, %10"  "\n\t" \
       "sub       %15, %11"  "\n\t" \
 \
       "mov       %8, %0"    "\n\t" \
       "mov       %9, %1"    "\n\t" \
       "mov       %10, %2"   "\n\t" \
       "mov       %11, %3"   "\n\t" \
       "sub       %17, %8"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %8, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %17, %9"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %9, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %17, %10"  "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %10, %2"   "\n"   \
       "2:"                  "\t"   \
       "sub       %17, %11"  "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %11, %3"   "\n"   \
       "3:"                         \
       : "=m" ((X)[8]), "=m" ((X)[9]), "=m" ((X)[10]), "=m" ((X)[11]), \
         "=&x" (x0), "=&x" (x1), "=&x" (x2), "=&x" (x3), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "x" (b_over_p), \
         "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), "m" ((X)[3]), \
         "m" (mod64_rnd) \
       : "cc" ); \
 \
  asm ("mov       %19, %8"   "\n\t" \
       "mov       %20, %9"   "\n\t" \
       "mov       %21, %10"  "\n\t" \
       "mov       %22, %11"  "\n\t" \
 \
       "cvtsi2sdq %8,  %4"   "\n\t" \
       "cvtsi2sdq %9,  %5"   "\n\t" \
       "cvtsi2sdq %10, %6"   "\n\t" \
       "cvtsi2sdq %11, %7"   "\n\t" \
       "mulsd     %18, %4"   "\n\t" \
       "mulsd     %18, %5"   "\n\t" \
       "mulsd     %18, %6"   "\n\t" \
       "mulsd     %18, %7"   "\n\t" \
       "imul      %16, %8"   "\n\t" \
       "imul      %16, %9"   "\n\t" \
       "imul      %16, %10"  "\n\t" \
       "imul      %16, %11"  "\n\t" \
       "cvtsd2siq %4, %12"   "\n\t" \
       "cvtsd2siq %5, %13"   "\n\t" \
       "cvtsd2siq %6, %14"   "\n\t" \
       "cvtsd2siq %7, %15"   "\n\t" \
 \
       "imul      %17, %12"  "\n\t" \
       "imul      %17, %13"  "\n\t" \
       "imul      %17, %14"  "\n\t" \
       "imul      %17, %15"  "\n\t" \
       "sub       %12, %8"   "\n\t" \
       "sub       %13, %9"   "\n\t" \
       "sub       %14, %10"  "\n\t" \
       "sub       %15, %11"  "\n\t" \
 \
       "mov       %8, %0"    "\n\t" \
       "mov       %9, %1"    "\n\t" \
       "mov       %10, %2"   "\n\t" \
       "mov       %11, %3"   "\n\t" \
       "sub       %17, %8"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %8, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %17, %9"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %9, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %17, %10"  "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %10, %2"   "\n"   \
       "2:"                  "\t"   \
       "sub       %17, %11"  "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %11, %3"   "\n"   \
       "3:"                         \
       : "=m" ((X)[12]), "=m" ((X)[13]), "=m" ((X)[14]), "=m" ((X)[15]), \
         "=&x" (x0), "=&x" (x1), "=&x" (x2), "=&x" (x3), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "x" (b_over_p), \
         "m" ((X)[4]), "m" ((X)[5]), "m" ((X)[6]), "m" ((X)[7]), \
         "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[8], b, p, where 0 <= X[i],b < p < 2^51:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 8.
   Assumes SSE operations round to zero.
   Assumes b_over_p = b/p computed in round-to-zero mode.
*/
#define VEC8_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  register double x0, x1, x2, x3; \
 \
  asm ("mov       %0, %8"    "\n\t" \
       "mov       %1, %9"    "\n\t" \
       "mov       %2, %10"   "\n\t" \
       "mov       %3, %11"   "\n\t" \
 \
       "cvtsi2sdq %8, %4"    "\n\t" \
       "cvtsi2sdq %9, %5"    "\n\t" \
       "cvtsi2sdq %10, %6"   "\n\t" \
       "cvtsi2sdq %11, %7"   "\n\t" \
       "mulsd     %18, %4"   "\n\t" \
       "mulsd     %18, %5"   "\n\t" \
       "mulsd     %18, %6"   "\n\t" \
       "mulsd     %18, %7"   "\n\t" \
       "imul      %16, %8"   "\n\t" \
       "imul      %16, %9"   "\n\t" \
       "imul      %16, %10"  "\n\t" \
       "imul      %16, %11"  "\n\t" \
       "cvtsd2siq %4, %12"   "\n\t" \
       "cvtsd2siq %5, %13"   "\n\t" \
       "cvtsd2siq %6, %14"   "\n\t" \
       "cvtsd2siq %7, %15"   "\n\t" \
 \
       "imul      %17, %12"  "\n\t" \
       "imul      %17, %13"  "\n\t" \
       "imul      %17, %14"  "\n\t" \
       "imul      %17, %15"  "\n\t" \
       "sub       %12, %8"   "\n\t" \
       "sub       %13, %9"   "\n\t" \
       "sub       %14, %10"  "\n\t" \
       "sub       %15, %11"  "\n\t" \
 \
       "mov       %8, %0"    "\n\t" \
       "mov       %9, %1"    "\n\t" \
       "mov       %10, %2"   "\n\t" \
       "mov       %11, %3"   "\n\t" \
       "sub       %17, %8"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %8, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %17, %9"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %9, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %17, %10"  "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %10, %2"   "\n"   \
       "2:"                  "\t"   \
       "sub       %17, %11"  "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %11, %3"   "\n"   \
       "3:"                         \
       : "+m" ((X)[0]), "+m" ((X)[1]), "+m" ((X)[2]), "+m" ((X)[3]), \
         "=&x" (x0), "=&x" (x1), "=&x" (x2), "=&x" (x3), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" (mod64_rnd) \
       : "cc" ); \
 \
  asm ("mov       %0, %8"    "\n\t" \
       "mov       %1, %9"    "\n\t" \
       "mov       %2, %10"   "\n\t" \
       "mov       %3, %11"   "\n\t" \
 \
       "cvtsi2sdq %8, %4"    "\n\t" \
       "cvtsi2sdq %9, %5"    "\n\t" \
       "cvtsi2sdq %10, %6"   "\n\t" \
       "cvtsi2sdq %11, %7"   "\n\t" \
       "mulsd     %18, %4"   "\n\t" \
       "mulsd     %18, %5"   "\n\t" \
       "mulsd     %18, %6"   "\n\t" \
       "mulsd     %18, %7"   "\n\t" \
       "imul      %16, %8"   "\n\t" \
       "imul      %16, %9"   "\n\t" \
       "imul      %16, %10"  "\n\t" \
       "imul      %16, %11"  "\n\t" \
       "cvtsd2siq %4, %12"   "\n\t" \
       "cvtsd2siq %5, %13"   "\n\t" \
       "cvtsd2siq %6, %14"   "\n\t" \
       "cvtsd2siq %7, %15"   "\n\t" \
 \
       "imul      %17, %12"  "\n\t" \
       "imul      %17, %13"  "\n\t" \
       "imul      %17, %14"  "\n\t" \
       "imul      %17, %15"  "\n\t" \
       "sub       %12, %8"   "\n\t" \
       "sub       %13, %9"   "\n\t" \
       "sub       %14, %10"  "\n\t" \
       "sub       %15, %11"  "\n\t" \
 \
       "mov       %8, %0"    "\n\t" \
       "mov       %9, %1"    "\n\t" \
       "mov       %10, %2"   "\n\t" \
       "mov       %11, %3"   "\n\t" \
       "sub       %17, %8"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %8, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %17, %9"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %9, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %17, %10"  "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %10, %2"   "\n"   \
       "2:"                  "\t"   \
       "sub       %17, %11"  "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %11, %3"   "\n"   \
       "3:"                         \
       : "+m" ((X)[4]), "+m" ((X)[5]), "+m" ((X)[6]), "+m" ((X)[7]), \
         "=&x" (x0), "=&x" (x1), "=&x" (x2), "=&x" (x3), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "x" (b_over_p), "m" (mod64_rnd) \
       : "cc" ); \
}
#endif /* defined(NEED_vector) && !defined(HAVE_vector) */
