/* sr5sieve.c -- (C) Geoffrey Reynolds, June 2006.

   Srsieve specialised for the Sierpinski/Riesel Base 5 projects.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <getopt.h>
#include "sr5sieve.h"
#include "version.h"
#include "arithmetic.h"

#if HAVE_MALLOPT
#include <malloc.h>
#endif

#define NAME BASE_NAME "sieve"
#if (BASE==0)
#define DESC "A sieve for multiple sequences k*b^n+/-1"
#else
#define DESC "A sieve for the Sierpinski/Riesel Base " XSTR(BASE) " projects"
#endif

/* One per process file names. */
char *checkpoint_file_name;
char *factors_file_name;
char *log_file_name;
static char *work_file_name;

#define DATA_FILE_NAME BASE_NAME "data.txt"
#define DEFAULT_CACHE_FILE_NAME BASE_NAME "cache.bin"

#if (BASE==5)
#define FACTORS_EMAIL "base5.factors@yahoo.com"
#endif

#if (BASE==2 || BASE==0)
#define SOB_DAT "SoB.dat"
#define RIESEL_DAT "riesel.dat"
#endif

int verbose_opt = 0;
uint32_t L1_cache_size = 0;
uint32_t L2_cache_size = 0;
#ifdef HAVE_vector
const char *baby_method_opt = NULL;
const char *giant_method_opt = NULL;
#endif
#if HASHTABLE_SIZE_OPT
uint32_t hashtable_size = 0;
#endif

static void banner(void)
{
  printf("%s %d.%d.%d -- %s.\n", NAME, MAJOR_VER, MINOR_VER, PATCH_VER, DESC);
}

static void help(void)
{
  printf("Usage: %s [OPTION ...]\n", NAME);
#if (BASE==5)
  printf(" -d --delete K         "
         "Delete the sequence K*%u^n+/-1 from `%s'.\n",
         (unsigned int)BASE, DATA_FILE_NAME);
  printf(" -g --newpgen K N0 N1  "
         "Generate a NewPGen file for K*%u^n+/-1 with N0 <= n < N1.\n",
         (unsigned int)BASE);
  printf(" -a --abc K N0 N1      "
         "Generate an ABC file for K*%u^n+/-1 with N0 <= n < N1.\n",
         (unsigned int)BASE);
#endif
#if (BASE==2 || BASE==0)
  printf(" -s --sierpinski       "
         "Sieve sequences k*2^n+1 in dat format file '%s'\n",SOB_DAT);
  printf(" -r --riesel           "
         "Sieve sequences k*2^n-1 in dat format file '%s'\n",RIESEL_DAT);
#endif
  printf(" -i --input FILE       "
         "Read sieve from FILE instead of `%s'.\n", DATA_FILE_NAME);
#if (BASE==2 || BASE==0)
  printf("                         Instead of `%s',`%s' when used with"
         " -s,-r.\n", SOB_DAT, RIESEL_DAT);
#endif
  printf(" -p --pmin P0          "
         "Sieve for factors p in the range P0 <= p <= P1 instead\n");
  printf(" -P --pmax P1          "
         " of reading ranges from work file `%s.%s'.\n",
         WORK_FILE_BASE, WORK_FILE_SUFFIX);
  printf(" -u --uid STRING       "
         "Append -STRING to base of per-process file names.\n");
#if LEGENDRE_CACHE
  printf(" -c --cache            "
         "Save Legendre symbol tables to `%s'.\n",DEFAULT_CACHE_FILE_NAME);
  printf(" -C --cache-file FILE  "
         "Load (or save) Legendre symbol tables from (or to) FILE.\n");
#endif
#if MULTI_PATH
  printf("    --sse2             "
         "Use SSE2 vector optimisations. (Default if supported).\n");
  printf("    --no-sse2          "
         "Don't use SSE2 vector optimisations.\n");
#endif
  printf(" -l --L1-cache SIZE    "
         "Assume L1 data cache is SIZE Kb.\n");
  printf(" -L --L2-cache SIZE    "
         "Assume L2 cache is SIZE Kb.\n");
#ifdef HAVE_vector
  printf(" -B --baby METHOD      "
         "Use METHOD for baby step mulmods.\n");
  printf(" -G --giant METHOD     "
         "Use METHOD for giant step mulmods.\n");
#endif
#if HASHTABLE_SIZE_OPT
  printf(" -H --hashtable SIZE   "
         "Force use of a SIZE Kb hashtable.\n");
#endif
#if SUBSEQ_Q_OPT
#if (BASE == 0)
  printf(" -Q --subseq Q         "
         "Force sieving k*b^n+c as subsequences (k*b^d)*(b^Q)^m+c.\n");
#else
  printf(" -Q --subseq Q         "
         "Force sieving k*%u^n+c as subsequences (k*%u^d)*(%u^Q)^m+c.\n",
         (unsigned int)BASE,(unsigned int)BASE,(unsigned int)BASE);
#endif
#endif
  printf(" -D --duplicates       "
         "Report factors that don't eliminate any composite.\n");
#if IDLE_OPT
#if IDLE_DEFAULT
  printf(" -z --idle             "
         "Run at idle priority. (Default)\n");
  printf(" -Z --no-idle          "
         "Don't adjust priority.\n");
#else
  printf(" -z --idle             "
         "Run at idle priority.\n");
  printf(" -Z --no-idle          "
         "Don't adjust priority. (Default)\n");
#endif
#endif
  printf(" -v --verbose          "
         "Print some extra messages. -vv prints more.\n");
  printf(" -h --help             "
         "Print this help.\n\n");

  printf("Without arguments (normal usage), read sieve from `%s',"
         " read ranges\n", DATA_FILE_NAME);
  printf("A,B from `%s.%s', sieve for factors p in the range"
         " A*10^9 <= p <= B*10^9.\n", WORK_FILE_BASE, WORK_FILE_SUFFIX);

  exit(EXIT_SUCCESS);
}

#if (BASE==5)
static const char *short_opts = "d:g:a:i:p:P:u:cC:A:l:L:B:G:H:Q:DzZvh";
#else
static const char *short_opts = "sri:p:P:u:cC:A:l:L:B:G:H:Q:DzZvh";
#endif
static const struct option long_opts[] =
  {
#if (BASE==5)
    {"delete",     required_argument, 0, 'd'},
    {"newpgen",    required_argument, 0, 'g'},
    {"abc",        required_argument, 0, 'a'},
#endif
#if (BASE==2 || BASE==0)
    {"sierpinski", required_argument, 0, 's'},
    {"riesel",     required_argument, 0, 'r'},
#endif
    {"input",      required_argument, 0, 'i'},
    {"pmin",       required_argument, 0, 'p'},
    {"pmax",       required_argument, 0, 'P'},
    {"uid",        required_argument, 0, 'u'},
#if LEGENDRE_CACHE
    {"cache",      no_argument,       0, 'c'},
    {"cache-file", required_argument, 0, 'C'},
#endif
#if MULTI_PATH
    {"no-sse2",    no_argument,       0, '1'},
    {"sse2",       no_argument,       0, '2'},
#endif
    {"L1-cache",   required_argument, 0, 'l'},
    {"L2-cache",   required_argument, 0, 'L'},
#ifdef HAVE_vector
    {"baby",       required_argument, 0, 'B'},
    {"giant",      required_argument, 0, 'G'},
#endif
#if HASHTABLE_SIZE_OPT
    {"hashtable",  required_argument, 0, 'H'},
#endif
#if SUBSEQ_Q_OPT
    {"subseq",     required_argument, 0, 'Q'},
#endif
    {"duplicates", no_argument,       0, 'D'},
#if IDLE_OPT
    {"idle",       no_argument,       0, 'z'},
    {"no-idle",    no_argument,       0, 'Z'},
#endif
    {"verbose",    no_argument,       0, 'v'},
    {"help",       no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };

static char *per_process_file_name(const char *base_name, const char *suffix,
                                   const char *uid)
{
  size_t len;
  char *name;

  len = strlen(base_name) + 1 + strlen(suffix) + 1;
  if (uid != NULL)
    len = len + strlen(uid) + 1;
  name = xmalloc(len);
  if (uid != NULL)
    sprintf(name,"%s-%s.%s",base_name,uid,suffix);
  else
    sprintf(name,"%s.%s",base_name,suffix);

  return name;
}

static time_t start_date;
static uint64_t work_pmin;
static uint32_t terms_read = 0;
static int read_work_file(const char *uid);
static void update_work_file(void);

static int opt_ind, opt_c;

static void attribute ((noreturn)) argerror(const char *str)
{
  if (long_opts[opt_ind].val == opt_c)
    error("--%s %s: argument %s.", long_opts[opt_ind].name, optarg, str);
  else
    error("-%c %s: argument %s.", opt_c, optarg, str);
}

static uint64_t parse_uint(uint64_t limit)
{
  uint64_t num;
  uint32_t expt;
  char *tail;

  errno = 0;
  num = strtoull(optarg,&tail,0);

  if (errno == 0 && num <= limit)
  {
    switch (*tail)
    {
      case 'e':
      case 'E':
        expt = strtoul(tail+1,&tail,0);
        if (errno != 0)
          goto range_error;
        if (*tail != '\0')
          break;
        while (expt-- > 0)
          if (num > limit/10)
            goto range_error;
          else
            num *= 10;
      case '\0':
        return num;
    }

    argerror("is malformed");
  }

 range_error:
  argerror("is out of range");
}

#if (__x86_64__ && MULTI_PATH)
static int x86_64_select_code_path(uint64_t pmax, int old_path)
{
  int new_path = old_path;

  if (old_path == 2 && pmax >= (UINT64_C(1) << 51))
  {
    if (verbose_opt)
      report(1,"Range end too high for SSE2 code path, switching to FPU.");
    fini_sieve_sse2();
    new_path = 1;
  }
  else if (old_path == 1 && pmax < (UINT64_C(1) << 51))
  {
    if (verbose_opt)
      report(1,"Range end low enough, switching to SSE2 code path.");
    fini_sieve();
    new_path = 2;
  }

  return new_path;
}
#endif

int main(int argc, char **argv)
{
  int want_help = 0;
  uint32_t i;
#if (BASE==5)
  uint32_t k_to_write = 0, k_to_delete = 0;
  uint32_t newpgen_opt = 0, abc_opt = 0, n0 = 0, n1 = UINT32_MAX;
#endif
#if (BASE==2 || BASE==0)
  int riesel_dat_opt = 0, sob_dat_opt = 0;
#endif
#if MULTI_PATH
  int code_path = 0;
#if __x86_64__
  int force_fpu = 0;
#endif
#endif
#if LEGENDRE_CACHE
  int legendre_cache_opt = 0, legendre_cache_file_opt = 0;
  const char *cache_file_name = DEFAULT_CACHE_FILE_NAME;
#endif
#if IDLE_OPT
  int idle_opt = IDLE_DEFAULT;
#endif
  const char *data_file_name = NULL, *uid = NULL;
  uint32_t L1_opt = 0, L2_opt = 0;

#if HAVE_MALLOPT
  /* All memory allocation takes place during initialization, so it doesn't
     have to be fast. Reducing this threshold allows more memory to be
     returned to the system when free() is called.
  */
  mallopt(M_MMAP_THRESHOLD,16000);
#endif

#if USE_COMMAND_LINE_FILE
  /* If no comand line arguments are given, attempt to read them from file.
   */
  if (argc == 1)
    read_argc_argv(&argc,&argv,NAME COMMAND_LINE_FILE_NAME_SUFFIX);
#endif

  while ((opt_c = getopt_long(argc,argv,short_opts,long_opts,&opt_ind)) != -1)
    switch (opt_c)
    {
#if (BASE==5)
      case 'd':
        k_to_delete = strtoul(optarg,NULL,0);
        break;
      case 'g':
        newpgen_opt = 1;
        k_to_write = strtoul(optarg,NULL,0);
        break;
      case 'a':
        abc_opt = 1;
        k_to_write = strtoul(optarg,NULL,0);
        break;
#endif
#if (BASE==2 || BASE==0)
      case 'r':
        riesel_dat_opt = 1;
        break;
      case 's':
        sob_dat_opt = 1;
        break;
#endif
      case 'i':
        data_file_name = optarg;
        break;
      case 'p':
        p_min = parse_uint(UINT64_MAX);
        break;
      case 'P':
        p_max = parse_uint(UINT64_MAX);
        break;
      case 'u':
        uid = optarg;
        break;
#if LEGENDRE_CACHE
      case 'c':
        legendre_cache_opt = 1;
        break;
      case 'C':
        legendre_cache_file_opt = 1;
        cache_file_name = optarg;
        break;
#endif
#if MULTI_PATH
      case '1':
        code_path = 1;
#if __x86_64__
        force_fpu = 1;
#endif
        break;
      case '2':
        code_path = 2;
#if __x86_64__
        force_fpu = 0;
#endif
        break;
#endif
      case 'l':
        L1_opt = strtoul(optarg,NULL,0);
        break;
      case 'L':
        L2_opt = strtoul(optarg,NULL,0);
        break;
#ifdef HAVE_vector
      case 'B':
        baby_method_opt = optarg;
        break;
      case 'G':
        giant_method_opt = optarg;
        break;
#endif
#if HASHTABLE_SIZE_OPT
      case 'H':
        i = strtoul(optarg,NULL,0) * 1024;
        for (hashtable_size = 1024; hashtable_size < i; )
          hashtable_size <<= 1;
        break;
#endif
#if SUBSEQ_Q_OPT
      case 'Q':
        subseq_Q = strtoul(optarg,NULL,0);
        break;
#endif
      case 'v':
        verbose_opt++;
        break;
      case 'D':
        duplicates_opt = 1;
        break;
#if IDLE_OPT
      case 'z':
        idle_opt = 1;
        break;
      case 'Z':
        idle_opt = 0;
        break;
#endif
      case 'h':
        want_help = 1;
        break;
      default:
        return 1;
    }

  banner();

  if (want_help)
    help();


  /* We install these signal handlers early because the default handler
     terminates the program. SIGTERM and SIGINT handlers are not installed
     until sieving begins, but termination is the right default for them.
  */
#ifdef SIGUSR1
  signal(SIGUSR1,handle_signal);
#endif
#ifdef SIGUSR2
  signal(SIGUSR2,handle_signal);
#endif

  log_file_name = per_process_file_name(LOG_FILE_BASE,LOG_FILE_SUFFIX,uid);

#if SUBSEQ_Q_OPT
  if (subseq_Q % BASE_MULTIPLE != 0)
    error("Subsequence base exponent Q=%"PRIu32" is not a multiple of %u.",
          subseq_Q, (unsigned int)BASE_MULTIPLE);
  if (subseq_Q != 0 && LIMIT_BASE % subseq_Q != 0)
    error("Subsequence base exponent Q=%"PRIu32" is not a divisor of %u.",
          subseq_Q, (unsigned int)LIMIT_BASE);
#endif

#if LEGENDRE_CACHE
  if (legendre_cache_opt && legendre_cache_file_opt)
    error("-c --cache and -C --cache-file switches are mutually exclusive.");
#endif

#if MULTI_PATH
  if (code_path == 0)
  {
#if __x86_64__
    /* x86-64 always has SSE2. Choose default based on p_max instead. */
    code_path = (p_max < (UINT64_C(1) << 51)) ? 2 : 1;
#else
    code_path = (have_sse2()) ? 2 : 1;
#endif
  }
  if (verbose_opt && code_path == 2)
    printf("Using SSE2 code path, ");
#endif /* MULTI_PATH */

  set_cache_sizes(L1_opt,L2_opt);


#if IDLE_OPT
  if (idle_opt)
    set_idle_priority();
#endif

#if (BASE==5)
  if (newpgen_opt && abc_opt)
    error("Switches -g and -a are mutually exclusive.");
  if (k_to_write)
  {
    if (optind < argc)
      n0 = strtoul(argv[optind++],NULL,0);
    if (optind < argc)
      n1 = strtoul(argv[optind++],NULL,0);
  }
#endif

  if (optind < argc)
    error("Unhandled argument %s.", argv[optind]);

  work_pmin = p_min;

#if (BASE==2 || BASE==0)
  if (sob_dat_opt && riesel_dat_opt && data_file_name != NULL)
    error("Not all three of -i -s -r can be used together.");
  if (sob_dat_opt)
  {
    uint32_t t;
    const char *file_name = (data_file_name) ? : SOB_DAT;
    report(0,"Reading `%s' ...",file_name);
    if ((t = read_dat_file(file_name,1)) == 0)
      error("Invalid or unrecognised file `%s'.",file_name);
    terms_read += t;
  }
  if (riesel_dat_opt)
  {
    uint32_t t;
    const char *file_name = (data_file_name) ? : RIESEL_DAT;
    report(0,"Reading `%s' ...",file_name);
    if ((t = read_dat_file(file_name,-1)) == 0)
      error("Invalid or unrecognised file `%s'.",file_name);
    terms_read += t;
  }
  if (!sob_dat_opt && !riesel_dat_opt)
  {
    const char *file_name = (data_file_name) ? : DATA_FILE_NAME;
    report(0,"Reading `%s' ...",file_name);
    if ((terms_read = read_abcd_file(file_name)) == 0)
      error("Invalid or unrecognised file `%s'.",file_name);
  }
#else
  data_file_name = (data_file_name) ? : DATA_FILE_NAME;
  report(0,"Reading `%s' ...", data_file_name);
  if ((terms_read = read_abcd_file(data_file_name)) == 0)
    error("Invalid or unrecognised file `%s'.", data_file_name);
#endif

#if (BASE==5)
  if (k_to_write)
  {
    for (i = 0; i < seq_count; i++)
      if (SEQ[i].k == k_to_write)
        break;
    if (i == seq_count)
      error("%"PRIu32"*%u^n+/-1 not found in %s.",
            k_to_write, (unsigned int)BASE, data_file_name);
    if (newpgen_opt)
      write_newpgen_k(i,n0,n1);
    else
      write_abc_k(i,n0,n1);
    exit(EXIT_SUCCESS);
  }

  if (k_to_delete)
  {
    overwrite_abcd(p_min,data_file_name,k_to_delete);
    exit(EXIT_SUCCESS);
  }
#endif

  finish_candidate_seqs();

#if LEGENDRE_CACHE
  if (legendre_cache_opt)
  {
    /* Force writing of a new cache file, even if it already exists.
     */
    generate_legendre_tables(NULL);
    write_legendre_cache(cache_file_name);
  }
  else
  {
    generate_legendre_tables(cache_file_name);
    if (legendre_cache_file_opt)
    {
      FILE *tmp;

      /* Write a new cache file only if one doesn't already exist.
       */
      if ((tmp = fopen(cache_file_name,"r")) == NULL)
        write_legendre_cache(cache_file_name);
      else
        fclose(tmp);
    }
  }
#else
  generate_legendre_tables(NULL);
#endif

  checkpoint_file_name =
    per_process_file_name(CHECKPOINT_FILE_BASE,CHECKPOINT_FILE_SUFFIX,uid);

  if (work_pmin || p_max) /* --pmin or --pmax given on the command line */
  {
    work_pmin = p_min; /* in case it was set from the input sieve file. */
    if (p_min<=k_max || p_min>=p_max || p_max>=(UINT64_C(1)<<MOD64_MAX_BITS))
      error("`-p --pmin P0' and `-P --pmax P1' must satisfy %"
            PRIu32" < P0 < P1 < 2^%u.", k_max, (unsigned int)MOD64_MAX_BITS);

    factors_file_name =
      per_process_file_name(FACTORS_FILE_BASE,FACTORS_FILE_SUFFIX,uid);
    p_min = read_checkpoint(p_min,p_max);
#if MULTI_PATH
#if __x86_64__
    if (!force_fpu)
      code_path = x86_64_select_code_path(p_max,code_path);
#endif
    switch (code_path)
    {
      default:
      case 1:
        sieve();
        break;
      case 2:
        sieve_sse2();
        break;
    }
#else
    sieve();
#endif
  }
  else /* --pmin, --pmax not given, read ranges from the work file. */
  {
    work_file_name =
      per_process_file_name(WORK_FILE_BASE,WORK_FILE_SUFFIX,uid);
    while (read_work_file(uid))
    {
      work_pmin = p_min;
      factor_count = 0;
      set_accumulated_cpu(0.0);
      p_min = read_checkpoint(p_min,p_max);
      notify_event(sieve_parameters_changed);
#if MULTI_PATH
#if __x86_64__
    if (!force_fpu)
      code_path = x86_64_select_code_path(p_max,code_path);
#endif
    switch (code_path)
    {
      default:
      case 1:
        sieve();
        break;
      case 2:
        sieve_sse2();
        break;
    }
#else
    sieve();
#endif
#ifdef FACTORS_EMAIL
      if (factor_count > 0)
        report(1,"Please send `%s' to %s.", factors_file_name, FACTORS_EMAIL);
#endif
      update_work_file();
    }
  }
#if MULTI_PATH
  switch (code_path)
  {
    default:
    case 1:
      fini_sieve();
      break;
    case 2:
      fini_sieve_sse2();
      break;
  }
#else
  fini_sieve();
#endif

  return EXIT_SUCCESS;
}

#define RANGE_BLOCK 1000000000
static int read_work_file(const char *uid)
{
  uint64_t p0, p1;
  char buf[100];
  FILE *work_file;
  int ret;

  report(0,"Reading project work file `%s' ...",work_file_name);
  work_file = xfopen(work_file_name,"r",error);

  ret = fscanf(work_file, "%"SCNu64"%*[-,\t ]%"SCNu64,&p0,&p1);
  xfclose(work_file,work_file_name);
  switch (ret)
  {
    case EOF:
      report(1,"No more work in `%s'.",work_file_name);
      return 0;

    case 2:
      if (p0 >= p1)
        error("Bad range %"PRIu64",%"PRIu64" in `%s'.",p0,p1,work_file_name);
      if (p1 > ((uint64_t)(((uint64_t)1 << MOD64_MAX_BITS)-1))/RANGE_BLOCK)
        error("Range end %"PRIu64" too high in `%s'.",p1,work_file_name);

      if (factors_file_name != NULL)
        free(factors_file_name);
      sprintf(buf,"%s%"PRIu64,FACTORS_FILE_BASE,p0);
      factors_file_name = per_process_file_name(buf,FACTORS_FILE_SUFFIX,uid);
      p_min = p0*RANGE_BLOCK;
      p_max = p1*RANGE_BLOCK;
      if (p_min <= k_max)
      {
        p_min = k_max+1;
        warning("Range start %"PRIu64" too low in `%s'.",p0,work_file_name);
        warning("Starting sieve at p=%"PRIu64".",p_min);
      }
      if (p_max <= p_min)
        error("Range end %"PRIu64" too low in `%s'.",p1,work_file_name);
      return 1;

    case 0:
    case 1:
      error("Malformed first line in `%s'.",work_file_name);

    default:
      error("Error while reading `%s'.",work_file_name);
  }
}

#define MAX_WORK_LINES 1000
#define MAX_LINE_LENGTH 80
static void update_work_file(void)
{
  FILE *work_file;
  char *lines[MAX_WORK_LINES];
  int i, j;

  work_file = xfopen(work_file_name,"r",error);

  for (i = 0; i < MAX_WORK_LINES; i++)
  {
    lines[i] = xmalloc(MAX_LINE_LENGTH);
    if (fgets(lines[i],MAX_LINE_LENGTH,work_file) == NULL)
    {
      free(lines[i]);
      break;
    }
  }

  xfclose(work_file,work_file_name);
  work_file = xfopen(work_file_name,"w",error);

  for (j = 1; j < i; j++)
    fputs(lines[j],work_file);
  xfclose(work_file,work_file_name);

  for (j = 0; j < i; j++)
    free(lines[j]);
}

/* Return the fraction of the current range that has been done.
 */
double frac_done(uint64_t p)
{
  return (double)(p-work_pmin)/(p_max-work_pmin);
}

#define STATUS_STRFTIME_FORMAT "ETA %d %b %H:%M"
void print_status(uint64_t p, uint32_t p_per_sec)
{
  static int toggle = 0;
  char buf[32];

  toggle = !toggle; /* toggle between reporting ETA and factor rate. */

  if (toggle && factor_count && p_per_sec)
  {
    uint64_t p_per_factor = (p-work_pmin)/factor_count;
    snprintf(buf,sizeof(buf),"%"PRIu64" sec/factor",p_per_factor/p_per_sec);
  }
  else
  {
    double done = (double)(p-p_min)/(p_max-p_min);
    if (done > 0.0) /* Avoid division by zero */
    {
      time_t finish_date = start_date+(time(NULL)-start_date)/done;
      strftime(buf,sizeof(buf),STATUS_STRFTIME_FORMAT,localtime(&finish_date));
    }
    else
      buf[0] = '\0';
  }

  report(0,"p=%"PRIu64", %"PRIu32" p/sec, %"PRIu32" factor%s, %.2f%% done, %s",
         p,p_per_sec,factor_count,plural(factor_count),100.0*frac_done(p),buf);
}

#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
#include <fenv.h>
static int old_mode;
#endif

static double expected_factors(uint32_t n, uint64_t p0, uint64_t p1)
{
  /* TODO: Use a more accurate formula. This one is only reasonable when
     p0/p1 is close to 1.
   */

  return n*(1-log(p0)/log(p1));
}

void start_srsieve(void)
{
  report(1,"Expecting to find factors for about %.2f terms in this range.",
           expected_factors(terms_read,work_pmin,p_max));

  logger("%s started: %"PRIu32" <= n <= %"PRIu32", %"PRIu64" <= p <= %"PRIu64,
         NAME, n_min, n_max, p_min, p_max);

#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
  old_mode = fegetround();
#endif

  start_date = time(NULL);
}

void finish_srsieve(const char *reason, uint64_t p)
{
#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
  int new_mode = fegetround();
  if (old_mode != new_mode)
    warning("FPU rounding mode was not restored: old=%d, new=%d.",
            old_mode, new_mode);
#endif

  logger("%s stopped: at p=%"PRIu64" because %s.", NAME, p, reason);
  logger("Found factors for %"PRIu32" term%s in %.3f cpu sec."
         " (expected about %.2f)",factor_count,plural(factor_count),
         get_accumulated_cpu(),expected_factors(terms_read,work_pmin,p));

#if REMOVE_CHECKPOINT
  if (p >= p_max) /* Sieving job completed */
    remove_checkpoint();
  else
#endif
    write_checkpoint(p);
}
